#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/uio.h>
#include <sys/types.h>

# define READ   0
# define WRITE  1

# define PARENT 0
# define CHILD  1

# define SUCCESS 0
# define FAILURE 1

# define STDIN  0
# define STDOUT 1

# define OPQ_CMD    "/home/kbre0/usr1/kbre/testversion/bin/query"
# define OPQ_CMD_ARG1 "-b"
# define OPQ_DB "/home/kbre0/usr4/kbre/dbs/full.mf"

# define PIPE(P,M) ((P) == PARENT ? pipePC[M] : pipeCP[M])

# define BUFLEN 500 /* buffer length for reads */

/* process id to wait for before finishing the parent */
static pid_t childpid = 0;


int
opq_write(FILE *pipeWrite, const char *cmd)
/**************************************************
 * sends `cmd' to the query module
 */
{
  fprintf(pipeWrite, "%s\n", cmd);
  fflush(pipeWrite);

  return(SUCCESS);
}


int
inStr( char c, char *str)
/**************************************************
 */
{
  for (; (c != *str) && (*str); str++) ;

  return(c == *str);
}

char *
extr_word(int *pos, char buf[])
/**************************************************
 * Extract the word starting at pos in the buffer buf.
 * The word ends at the end of buf or on a delimiter
 * delimiters may be spaces or quotes.
 * There is a bug with quoted strings containing their own quote inside.
 * for example:  'a string with '' inside'
 */
{
  int  start_word, word_length;
  char *res;
  char word[BUFLEN];
  char *space = "\t ",         /* char marking end of a word  */
	   *quote = "\"'",         /* char marking quoted-strings */
	    str1[2],               /* a 1-char string (+ '\0')    */
       *endwrd;                /* one of the two precedings   */

  str1[1] = '\0';

  while (buf[*pos] != '\0' && (buf[*pos] == ' ' || buf[*pos] == '\t'))
	(*pos)++;

  start_word = *pos;

  if ( inStr(buf[*pos],quote) ) {
	/* end of the word marked by matching string */
	str1[0] = buf[*pos];
	endwrd = str1;
	(*pos)++;
  }
  else
	endwrd = space;

  while ( (! inStr(buf[*pos],endwrd)) && (buf[*pos] != '\0') )
	(*pos)++;

  word_length = *pos - start_word;
  if (inStr(buf[*pos],quote) ) {
	 word_length++;   /* we want to put the closing quote in the word */
	 (*pos)++;        /* pass the end of the word (for next time) */
  }

  res = (char*)malloc(word_length + 1);
  strncpy(res, buf + start_word, word_length);
  res[word_length] = '\0';
  
  return(res);
}


int
opq_read( FILE *pipeRead, int *id, char **so, char **cls)
/**************************************************
 * reads answer from query module to a command.
 * answer takes the form: `SoftwareObject ID  Class VirtualClass'
 * respectively returned in the 3rd, 2nd & 4th arguments
 * the Virtual class (4th element) is not used.  It may also be an error
 *  message if there is no VirtualClass (i.e. ManifestConstantExistence)
 * if there is no answer to the query, returns -1 and two empty strings
 */
{
  char  line[BUFLEN]; /* a buffer to read the answert (line by line) */
  char  temp[BUFLEN];
  char  c;
  char *l;     /* pointer inside `line' */
  int b_pos;

  for( l=line, c=getc(pipeRead); c != '\n'; l++) {
	*l = c;
	c=getc(pipeRead);
  }

  *l = '\0';    /* terminating the `line' string */

  if (line[0] == '\0') {
	/* empty answer: software object id = -1 */
	*id = -1;
	/* empty answer: softwareObject = className = empty string */
	*so = (char*)malloc(1);
	*so[0] = '\0';
	*cls = *so;
  }
  else {
	b_pos = 0;

	/* "reading" the first atom (the softwareObject) inside `line' */
	strcpy(temp,extr_word(&b_pos,line));
	*so = (char *)malloc(strlen(temp)+1); 
	strcpy(*so,temp);

	/* reading the key */
	strcpy(temp,extr_word(&b_pos,line));
	*id = atoi(temp);

	/* "reading" the class name */
	strcpy(temp,extr_word(&b_pos,line));
	*cls = (char *)malloc(strlen(temp)+1); 
	strcpy(*cls,temp);

	/* we are not using the virtual class name which comes last in the line */
	/* strcpy(temp,extr_word(&b_pos,line)); */
  }

  return(SUCCESS);
}


int
opq_popen(char *cmd, char *DB, FILE **pipeRead, FILE **pipeWrite)
/**************************************************
 */
{
  int pipePC[2],   /* pipe PARENT -> CHILD  */
	  pipeCP[2];   /* pipe CHILD  -> PARENT */
  int rc;          /* return code of some system calls */
  pid_t pid;
  char c, *p;      /* used in the parent part */
  int   voidInt;
  char *voidPtr1, *voidPtr2;

  *pipeRead  = NULL;
  *pipeWrite = NULL;

  if ( (rc=pipe(pipePC)) != 0) {
	fprintf(stderr,"** Could not open pipe P->C, rc=%d\n", rc);
	return(FAILURE);
  }

  if ( (rc=pipe(pipeCP)) != 0) {
	fprintf(stderr,"** Could not open pipe C->P, rc=%d\n", rc);
	return(FAILURE);
  }

  if ( (pid=fork()) == (pid_t)-1) {
	fprintf(stderr,"** Could not fork\n");
	return(FAILURE);
  }

  /* ---------------------------------------------------------------------- */
  if (pid == (pid_t)0) {    /* child process */
	close( PIPE(CHILD, READ) );
	close( PIPE(PARENT,WRITE) );

	/* re-directing stdin from the pipe P->C */
	if ( dup2(PIPE(PARENT,READ), STDIN)  < 0 ) {
	  fprintf(stderr,"** child: Couldn't reopen stdin\n");
	  return(FAILURE);
	}
 	close( PIPE(PARENT,READ) );  /* no longer needed after the `dup' */

	/* re-directing stdout to the pipe C->P */
	if ( dup2(PIPE(CHILD,WRITE), STDOUT) < 0 ) {
	  fprintf(stderr,"** child: Couldn't reopen stdout\n");
	  return(FAILURE);
	}
 	close( PIPE(CHILD,WRITE) );  /* no longer needed after the `dup' */

	/* and calling the command */
	if ( (cmd == NULL) || (*cmd == '\0') )
	  cmd = OPQ_CMD;
	if ( (DB == NULL) || (*DB == '\0') )
	  DB = OPQ_DB;
 	execl(cmd, "query-piped", OPQ_CMD_ARG1, DB, (char *)0);

	fprintf(stderr,"** opq_popen: exec failed in child\n");
	return(FAILURE);
  }

  /* ---------------------------------------------------------------------- */
  else {
	/* parent process */
	childpid = pid;      /* used when closing the pipes */

	close( PIPE(PARENT,READ) );
	*pipeWrite = fdopen( PIPE(PARENT,WRITE), "w");
	if ( *pipeWrite == NULL ) {
	  fprintf(stderr,"** opq_popen: parent couldn't create write FILE descriptor\n");
	  return(FAILURE);
	}

	close( PIPE(CHILD,WRITE) );
	*pipeRead = fdopen( PIPE(CHILD,READ), "r");
	if ( *pipeRead == NULL ) {
	  fprintf(stderr,"** opq_popen: parent couldn't create read FILE descriptor\n");
	  return(FAILURE);
	}

	/* init the query module */
	if (opq_write(*pipeWrite, "set display = xcdbmode") != SUCCESS)
	  return(FAILURE);

	/* reads first line printed (a prompt and the answer to set display) */
	while ( fgetc(*pipeRead) != '\n' );
	/* reads the blankline after this first line */
	if (  fgetc(*pipeRead) != '\n' ) {
	  fprintf(stderr,"** opq_popen: parent doesn't understand OPQ protocole (set display=xcdbmode)\n");
	  return(FAILURE);
	}

	return(SUCCESS);
  }
}


int
opq_close(FILE *pipeRead, FILE *pipeWrite)
/**************************************************
 */
{
  opq_write( pipeWrite, "quit");
  fclose( pipeWrite);

  while (wait(0) != childpid) ;
  fclose( pipeRead);

  return(SUCCESS);
}
