
:- file_search_path(opqlib,_) ;
   assert(file_search_path( opqlib, '/home/kbre0/usr1/kbre/testversion/lib')).


foreign_file(opqlib(opq_pl), [opq_popen, opq_write, opq_read, opq_close]).

foreign(opq_popen,c,opq_popen([-integer],+string,-address('FILE'),
	                                             -address('FILE'))).
foreign(opq_write,c,opq_write([-integer],+address('FILE'),+string)).
foreign(opq_read, c,opq_read( [-integer],+address('FILE'),
	                                     -integer,-string,-string)).
foreign(opq_close,c,opq_close([-integer],+address('FILE'),+address('FILE'))).

:- load_foreign_files([opqlib(opq_pl)],[]).

opq_help :-
	format("Low level predicats (`R'/`W' are read/write channels):~n",_),
	format("- opq_popen([-bool],+PGM,-R,-W)To open a connection with the DB~n",_),
	format("- opq_write([-bool],+W,+Cmd)   Write a Cmd to the connection~n",_),
	format("- opq_read([-bool],+R,-ID,-SO,-CL) To get a SoftObj, its Key and its CLass from previously~n                               asked query.  ID=-1 means no (more) answer to~n                               the query~n",_),
	format("- opq_close([-bool],+R,+W)     To close the connection with the DB~n",_),
	format("~nHigher level predicats:~n",_),
	format("- opq_init/opq_init(+PGM)      Initialize opq connection~n",_),
	format("- opq_query(+Query)            To pass a query to the DB~n",_),
	format("- opq_answer(?ID,?SO,?CL)      Gets an answer to last opq_query, fails if~n                               no more answers~n",_),
	format("- opq_findall(+Query,?All)     Gets all answers to Query, `All' is a list of~n                               `so(ID,SO,Class)'~n",_),
	format("- opq_end                      Terminates a connection~n",_),
	format("- opq_clean                    Tries to cleanup a connection (after error)~n",_),
	format("- opq_opened                   Was OPQ initialized? (yes answer not reliable)~n",_),
	format("~n- Predicates with [-bool] are C functions returning 0 in case of succes~n- and a non nul integer otherwise.~n",_).

%% tests:
%% opq_popen(0,'',R,W),opq_write(0,W,'id 0'),opq_read(0,R,ID,SO,CL),opq_close(0,R,W).
%% opq_init, opq_findall('name process_id', All), opq_end.


:- dynamic opqStreams/2,  %% opqStreams(ReadPipe,WritePipe)
	       opqFindall/1.  %% local


opq_init :-
	opq_init('/home/kbre0/usr1/kbre/testversion/bin/query').

opq_init(_) :-
	opqStreams(_,_).
opq_init(PGM) :-
	\+ opqStreams(_,_),
	opq_popen(0,PGM,OPQ_Read,OPQ_Write),   %% `0' means SUCCESS
	assert( opqStreams(OPQ_Read,OPQ_Write)).


opq_query(QUser) :-
	'OPQ_concat'("first 0 ", QUser, Query),
	opqStreams(OPQ_Read,OPQ_Write),
	opq_write(0,OPQ_Write,Query),
	opq_read(0,OPQ_Read,-1,_, _).


opq_answer(ID,SO,CL) :-
	opqStreams(OPQ_Read,OPQ_Write),
	opq_write(0,OPQ_Write,'next 1'),

	opq_read(0,OPQ_Read,ID,SO,CL),
	SO \== '',
	opq_read(0,OPQ_Read,-1,_, _).


%% TRICKY PREDICATE: if everything goes normally, opq_read(0,...)
%% will "always" succeed, therefore opq_findall_assert_ should also
%% succeed, therefore we stay in the repeat loop until the test
%% ID=-1 is true (ID=-1 is the last answer we can get through
%% opq_read).  Upon this last answer, we try to remove the last
%% asserted opqFindall(so(-1,'',_)) which should succeed because we did
%% assert an opqFindall(so(-1,'',_))
%% On the other hand, if opq_read(0,...) fails, opq_findall_assert_ will
%% also fail, which stops the repeat loop and goes to the retract
%% which should also fail because we did not assert an
%% opqFindall(so(-1,'',_)).  Therefore, if opq_read(0,...) fails, 
%% opq_findall will also fail which seems a good idea
opq_findall(Query,All) :-
	opqStreams(OPQ_Read,OPQ_Write),
	opq_write(0,OPQ_Write,Query),

	retractall(opqFindall(_)),
	repeat,
	opq_findall_assert_(OPQ_Read,ID),
	ID = -1,
	!,
	retract(opqFindall(so(-1,'',_))),  %% remove the last assert: so(-1,'',_)

	findall(ID_SO_CL, opqFindall(ID_SO_CL), All).

opq_findall_assert_(OPQ_Read,ID) :-
	opq_read(0,OPQ_Read,ID,SO,CL),
	assert(opqFindall(so(ID,SO,CL))).


opq_end :-
	opqStreams(OPQ_Read,OPQ_Write),
	opq_close(0,OPQ_Read,OPQ_Write),
	retractall(opqStreams(_,_)).


opq_clean :-
	opq_end,
	!.
opq_clean :-   %% in case `opq_end' failed
	retractall(opqStreams(_,_)),
	!.
opq_clean.     %% in case `retractall' failed


'OPQ_concat'(Str, Atom, AtomOut) :-
	name( Atom, Str2),
	'OPQ_append'( Str, Str2, StrOut),
	name( AtomOut, StrOut).

'OPQ_append'( [], L2, L2).
'OPQ_append'( [X|L1], L2, [X|LOut]) :-
	'OPQ_append'( L1, L2, LOut).
