//                              -*- Mode: C++ -*- 
// Q-element.cc --- Query-element: internal representation for queries' results
// Author          : Nicolas Anquetil
// Created On      : Mon Aug 18 10:45:46 1997
// 

# include "Q-element.hh"


bool
Qelement::compatible( char *str, class_t clss)
// --------------------------------------------------
{
  return( (strcmp(qName(),str) == 0) &&
		   compatibleclass(qClass(),clss) );
}

bool
Qelement::compatible(SoftwareObject *so) 
  /********************************************************************
   * if the Qelements class and the SoftwareObjects type (and name)are 
   * compatible and  the SoftwareObject corresponds to a *missing*
   * component of the Qelement, then this SoftwareObject is
   * compatible with the Qelement
   */

{
  return( ((compatibleclass(so->classType() , qClass())) 
	  &&  (strcmp(qName(),so->objectName()) == 0)) 
	  &&  ((! so->is_a(_referenceexistence) && (defKey() == NULLKEY))
	       ||  ((so->is_a(_referenceexistence)) && (useKey() == NULLKEY ))) );
}




bool
Qelement::compatible( Qelement *qelt)
// --------------------------------------------------
{
  return(compatible( qelt->qName(), qelt->qClass()));
}

 
void
Qelement::merge( SoftwareObject *obj)
  //---------------------------------------------------------------
  // Assumes that the softwareObject 'obj' tested to be "compatible" 
  // with this Qelement. 
  // This method modifies this Q-element by the addition of a useKey 
  // value or a defKey value.  If it is a defKey being added then we 
  // also switch the Qelements qClass to the 'classType' of 'obj'  
{
  if (obj->is_a(_referenceexistence))  
      _useKey = obj->objectKey();
      else {
	_defKey = obj->objectKey();
	_qClass = obj->classType();
      }
}
      

void
Qelement::merge( Qelement *qelt)
// --------------------------------------------------
// May modify `this' Q-element
{
  if (qClass() != qelt->qClass())
	_qClass = virtualizeclass(_qClass);
}


VoidIterator<SoftwareObject*> *
Qelement::getObjectList()
// --------------------------------------------------
{
  VoidIterator<SoftwareObject*> *iter;
 
  switch (qClass()) {  
  case _file:
{    iter = new SortedUnionIterator<SoftwareObject*>(
			 cdbGetObjectListGivenName( qName(), _sourcefile),
			 cdbGetObjectListGivenName( qName(), _fileinclusionexistence),
			 SO_strict_cmp, SOmerge);
}
    break;

  case _routine:
{    iter = new SortedUnionIterator<SoftwareObject*>(
			 cdbGetObjectListGivenName( qName(), _routinesource),
			 cdbGetObjectListGivenName( qName(), _routinecallexistence),
			 SO_strict_cmp, SOmerge);

}
   break;

  case _type:
{    iter = new SortedUnionIterator<SoftwareObject*>(
			 cdbGetObjectListGivenName( qName(), _typedef),
			 cdbGetObjectListGivenName( qName(), _typeuseexistence),
			 SO_strict_cmp, SOmerge);

}    
break;

  case _data:
{
    iter = new SortedUnionIterator<SoftwareObject*>(
			 new SortedUnionIterator<SoftwareObject*>(
	                     new SortedUnionIterator<SoftwareObject*>(
                 cdbGetObjectListGivenName( qName(), _enumerationconst),
				 cdbGetObjectListGivenName( qName(), _field),
				 SO_strict_cmp, SOmerge),
			   cdbGetObjectListGivenName( qName(), _datumdef),
			   SO_strict_cmp, SOmerge),
			 cdbGetObjectListGivenName( qName(), _datauseexistence),
			 SO_strict_cmp, SOmerge);
} 
   break;

  default:
    iter = cdbGetObjectListGivenName( qName(), qClass());
    break;
  }

  return (iter);
}  // getObjectList

VoidIterator<SoftwareObject*> *
Qelement::getObjectList(class_t filter)
// --------------------------------------------------
{

  return( cdbGetObjectListGivenName( qName(), filter));

} // getObjectList()

SoftwareObject*
Qelement::getDefObject()
// --------------------------------------------------
{
 return( cdbGetObjectGivenKey(defKey()));
} 

SoftwareObject*
Qelement::getUseObject()
// --------------------------------------------------
{
 return( cdbGetObjectGivenKey(useKey()));
} 

void
Qelement::print()
// --------------------------------------------------
{

  if (displayMode == _xcdbmode) {

    if (_defKey != NULLKEY) 
      printf("%s %d %s %s\n", 
	     qName(), defKey(), cdbClass2str(qClass()), cdbClass2str(virtualizeclass(qClass()))  );
    else
      printf("%s %d %s %s\n", 
	     qName(), useKey(), cdbClass2str(qClass()), cdbClass2str(virtualizeclass(qClass())) );
    
  }
  
  else {
    if ( ( _defKey != NULLKEY ) && ( _useKey != NULLKEY ) ) {
      printf("%d:", defKey());   // print the defKey if present
      _qClass = virtualizeclass(_qClass) ;  // generalize the class
      if (displayMode == _dsplonger)
	printf("%d:", useKey());//print the useKey if _dsplonger mode
    } // both keys  present
    
    else { // a key is missing
      if ( _defKey == NULLKEY ) // if its the defKey missing
	printf("%d:", useKey()); // print the useKey  
      else // its the useKey thats missing
	printf("%d:", defKey()); // print the defKey  
    }   
    
    
    // Jerry Chen -- use write instead of printf so it will write to
    // pipe properly
    
    /*    char buff[80];
	  sprintf (buff, "%s\n", qName());
	  write (1, buff, strlen (buff));
	  */    
    printf("%s", qName());
    if ((displayMode == _dsplong) || (displayMode == _dsplonger))
      printf(" :%s\n", cdbClass2str(qClass())  );
    else
      printf(" :%s\n", cdbClass2abbrev(qClass()) );
    
  } //else not xcdbmode

}  // Qelement::print()

bool      
compatibleSO(SoftwareObject *x, SoftwareObject *y)
  /* Two software objects are compatible if they represent two aspects of
   * the same virtual object.  This means that they have the same name
   * and compatible class, and that one is the Definition while the other 
   * is the UseExistence aspect of the same virtual object.
   */
{
  return(((x->objectName() ==  y->objectName())
	  &&
	  (compatibleclass(x->classType(), y->classType()) ) )
	 &&  
	 ( (x->is_a(_referenceexistence)) != ( y->is_a(_referenceexistence)))
	 );
}






qeltCmp(Qelement *x, Qelement *y)
  // --------------------------------------------------
  // tests comptability rather than equality  
  // If the Qelements x and y are incompatible on the basis 
  // of a conflict in qName, qClass or defKey then we determine
  // which should precede the other in a list.
  // return value of 0 means they are compatible 
  // return value of -1 means x should precede y
  // return value of 1 means y should precede x

{
  char *xname, *yname;       // the qNames of x and y
  int   namecmp;             // holds result of strcmp on names
  int   xdefKey, ydefKey;    // hold the defKey values


  xname = x->qName();
  yname = y->qName();
  xdefKey = x->defKey();
  ydefKey = y->defKey();

  if ( (xname != NULL) && (yname != NULL) ) {
    /* if the names are both non-null */
    namecmp = strcmp(xname, yname);
    if      (namecmp != 0)
      return( namecmp);  // different names; order by name
    else if (!(compatibleclass(x->qClass(), y->qClass()))) { 
      /* different virtual class;  order by qClass value */
      if ((int)x->qClass() < (int)y->qClass())
	return( -1);  
      else
	return( 1);
    }
    else if (( (xdefKey != NULLKEY) && (ydefKey != NULLKEY) )
	     && (xdefKey != ydefKey)) {
      /* keys are non-null and different; order by defKey */
      if (xdefKey < ydefKey)
	return( -1);
      else
	return( 1);      
    }
    else return( 0); /* are "equal", i.e. compatible */
  }
  else {
    fprintf(stderr,"** Comparing Qelements with no name\n");
    exit(2);
  }
}  /* qeltCmp */


Qelement *
qeltMerge(Qelement *x, Qelement *y)
// --------------------------------------------------
// Merging two compatible Q-elements (=> they refer to the same 
// virtual object)
{
  Qelement *tmp;

  if (x->defKey() == NULLKEY) {
    tmp = new Qelement(*y);
    tmp->merge(x->getUseObject());
  }
  else {
    tmp = new Qelement(*x);
    tmp->merge(y->getDefObject());
  }
  return(tmp);
}   











