
First the grammar:

 1 <expr> ::=   {first {<n>}} <query>
 2            | next {<n>}
 3            | objinfo <objid>
 4            | classinfo <class>
 5            | sysinfo
 6            | set <option> = <argument>
 7 
 8 <query> ::=   id <objid>
 9             | {class} name <name>
10             | {class} reg <regexp>
11             | {class} pat <pattern>
12             | {class} {association} <query>
13             | <query> <operator> <query>
14             | ( <query> )
15
16 <class> ::=   <regularclass>
17             | <specialclass>

Explanations:

<n>
  - any integer greater than zero

<objid>
  - An integer used to represent an object in the database. Every real
    object in the database has an object id. This can be reliably used to
    refer to the same object repeatedly. Objects can also be referred to
    by name, but several objects may have the same name, and some objects
    may not have a name.

<name>
  - Any character string without whitespace, or else a character string
    enclosed in "quotes". 

<regexp>
  - Any grep-like regular expression without whitespace, or else a
    character string enclosed in "quotes".

<pattern>
  - A unix-like pattern such containing * as a wildcard.

<operator>
    and - set intersection
    or  - set union
    -   - set difference (elements of first set less elements of second set)

<regularclass>
  - Represents a class in the database. Full names of classes can be used,
    or else abbreviations, as listed here. (It is possible that this list
    of classes could be added-to or otherwise change).

    so   softwareObject

    su   sourceUnit
    sf   sourceFile     
    swf  SourceWithinFile
    rs   RoutineSource

    de   definition
    /* note that the following all have abbreviations that end with d */
    ec   enumerationConst
    tdd  typedDefinition
    sd   standaloneDefinition
    td   typeDef
    etd  enumeratedTypeDef
    rtd  recordTypeDef
    dd   datumDef
    fld  field

    /* note that the following all have abbreviations that end with e */
    re   referenceExistence
    fie  fileInclusionExistence
    due  dataUseExistence
    rce  routineCallExistence
    mce  manifestConstantExistence
    tue  typeUseExistence
    vue  variableUseExistence
    sce  specialCodeExistence

<specialclass>
    - Special classes, that are not really part of the database, but
       allow us to have virtual multiple inheritance. Others may be added.

    rou  routine
         - when requesting all named items that are mentioned in rs and rce
    fil  file
         - when requesting all named items that are mentioned in sf and fie
    dat  data
         - when requesting all named items that are mentioned in tdd and due
    typ  type
         - when requesting all named items that are mentioned in tf and tue

<association>
  - thatdef      
      <defclass> def <defobjs>
         -> return instances of defclass that define defobjs
         -> defclass can be sourceUnit, definition or any subclasses
         -> defobjs will be instances of these classes
         -> e.g.
            files that define routines
            sourceUnits that define standaloneDefinitions
            recordTypeDefinitions that define fields
            enumeratedTypeDefinitions that define enumerationConstants 

  - defby
      <defclass> defin <defobjs>
         -> inverses of the above
         -> e.g.
            routines defined in specific files
            standaloneDefinitions defined in specific sourceUnits
            fields defined in specific recordTypeDefinitions
            enumerationConstants defined in
                        sepcific enumerationTypeDefinitions 

  - thatref
      refclass ref refobjs
         -> return instances of refclass that refer to refobjs
         -> e.g. 
            routines that call specific routines
            files that include specific files
            sourceUnits that refer to specific referenceExistences
            routines that have specific typeUseExistences as return type
            typedDefinitions that have specific typeUseExistences as type

  - refby
      refclass refby refobjs
         -> inverse of the above
         -> e.g.
            routines called by specific routines
            files included by specific files
            referenceExistences that are refered to by specific sourceUnits
            typeUseExistences that are the return type of specific routines
            typeUseExistences that are the type of specific typedDefinitions

  Examples

      rs thatref (rce name 'xxx')
          -> return the routines that call routines with name 'xxx'

      rou thatref (rou name 'xxx')
          -> same as above, but less efficient. The nested query finds
             all defined routines and routine call existences that have
             the given name


<query>

  - always returns output as a list of objects, one per line.
    The default format is as follows:

      <class> <name> {<objid>}

  - in this case, class is always the 2 or 3-character abbreviation
  - objid *may be* omitted by the system.
  - <class> will always be a subclass of the original query class, or the
    original query class itself.
  - if there is no name, a ? is returned
  - the list is always terminated by an empty line containing 'end'


first {<n>} <query>
  - computes and returns only the first <n>, (or a default number) 
    of responses to the query.
  - leaves an open iterator in the database, if there are remaining
    answers.

next {n}
  - computes the next elements of the query, if there the last query
    used 'first', and if there are remaining answers.
  - returns an empty set otherwise

set
  - details to be defined
  - will be used to change the database etc.

The query language terminates on end-of-file.


The following table shows the membership of each "specialclass"
broken down by definitions and references:

+-----+-----------+------------+-----+-----+
|     |   dat     |    typ     | fil | rou |
+-----+-----------+------------+-----+-----+
| def | ec/dd/fld | td/rtd/etd | sf  | rs  |
+-----+-----------+------------+-----+-----+
| ref | due       | tue        | fie | rce |
+-----+-----------+------------+-----+-----+



Timothy C. Lethbridge,         | Assistant Professor / Professeur adjoint
http://www.csi.uottawa.ca/~tcl | tcl@site.uottawa.ca /T.Lethbridge@ieee.ca
SITE: School of Information Technology and Engineering
University of/d'Ottawa         | W/B: 613 562-5800 x6685
Ottawa, Canada, K1N 6N5        | Fax: 613 562-5187  H/M: 613 237-6642
