//                               -*- Mode: C++ -*- 
// api.cc  ---     : API between query language and DB
// Author          : Javed Chowdhury.
// Created On      : 
//

# include "api.hh"

void
sysinfo()
// --------------------------------------------------
{
  printOptions();
  printDBinfo();
  printf("\n");
   
}


void
printDBinfo()
// --------------------------------------------------
{
  printf("Database:\n");

  printf("version: %s\n", cdbVersion());
  printf("Number of Objects in the Database: %d\n",
		 classObjCount(_softwareobject));
  printf("Number of Definition in the Database: %d\n",
		 classObjCount(_definition));
  printf("Number of Reference Existence in the Database: %d\n",
		 classObjCount(_referenceexistence));
  printf("Number of Routine in the Database: %d\n", classObjCount(_routine));
  printf("Number of File in the Database: %d\n",    classObjCount(_file));
  printf("Number of Data in the Database: %d\n",    classObjCount(_data));
  printf("Number of Type in the Database: %d\n",    classObjCount(_type));
}   /* printDBinfo() */


void
classinfo(class_t clss)
// --------------------------------------------------
{
  printf("Number of %s: %d\n", cdbClass2str(clss), classObjCount(clss));
}


int
classObjCount(class_t clss)
// --------------------------------------------------
{
  int total;

  switch( clss) {
  case _routine:
    total = cdbGetObjectCount(_routinecallexistence) + 
            cdbGetObjectCount(_routinesource);
    break;
  case _file:
    total = cdbGetObjectCount(_sourcefile) + 
            cdbGetObjectCount(_fileinclusionexistence);
    break;
  case _data:
    total = cdbGetObjectCount(_datauseexistence) + 
            cdbGetObjectCount(_datumdef);
    break;
  case _type: 
    total = cdbGetObjectCount(_typeuseexistence) + 
            cdbGetObjectCount(_typedef);
    break;   
  default:
    total =  cdbGetObjectCount(clss);
    break;
  }

  return( total);
}   /* classObjCount() */


/*****************************************************
           functions related to OBJINFO
 *****************************************************/


void
objinfo(int key)
// --------------------------------------------------
{
  SoftwareObject *object;
  class_t clss;
    
  if ( (key >= cdbGetObjectCount(_softwareobject)) || (key < 0) )
    errorMsg("wrong object key !");
  else {
    object = cdbGetObjectGivenKey(key);
    clss = cdbGetClassGivenObject(object);
    
    switch(clss) {
    case _sourcefile:
      objectSF((SourceFile*)object);
      break;
    case _routinesource:
      objectRS((RoutineSource*)object);
      break;
    case _fileinclusionexistence:
      objectFIE((FileInclusionExistence*)object);
      break;
    case _manifestconstexistence:
      objectMCE((ManifestConstExistence*)object);
      break;
    case _datauseexistence:
      objectDUE((DataUseExistence*)object);
      break;
    case _typeuseexistence:
      objectTUE((TypeUseExistence*)object);
      break;
    case _routinecallexistence:
      objectRCE((RoutineCallExistence*)object);
      break;
    case _field:
      objectFLD((Field*)object);
      break;
    case _datumdef:
      objectDD((DatumDef*)object);
      break;
    case _enumerationconst:
      objectEC((EnumerationConst*)object);
      break;
    case _recordtypedef:
      objectRTD((RecordTypeDef*)object);
      break;
    case _enumeratedtypedef:
      objectETD((EnumeratedTypeDef*)object);
      break;
    }
  }
}   /* objinfo() */

void
objectSF(SourceFile *object)
// --------------------------------------------------
{
  char fileName[MAX_STR_SIZE],
	   path[MAX_STR_SIZE];
  int  version,
	   dateChanged,
	   size;
  VoidIterator<SoftwareObject*> *itIncludes;

  cdbGetSourceFileAttributes(object, fileName, path, &version, 
			     &dateChanged, &size, &itIncludes);
  if (displayMode == _dsplong)
    printf("SourceFile: %s, path: %s, version: %d, dateChange: %d, size: %d\n",
	   fileName, path, version, dateChanged, size);
  else if (displayMode == _xcdbmode)
    printf("%s %d %d %s %d %d",
	   fileName, object->objectKey(), size, path, version, dateChanged);
  else
    printf("%s %s\n", fileName, path);
}   /* objectSF() */



void
objectRS(RoutineSource *object)
// --------------------------------------------------
{
  char *pcdFcn;
  SourceUnit *contSource;
  int sChar,
	  eChar;
  char routineName[MAX_STR_SIZE];
  bool isEntry;
  TypeUseExistence *retType;
  VoidIterator<SoftwareObject*> *itCalls;
  char defby[MAX_STR_SIZE];

  cdbGetRoutineSourceAttributes(object, routineName, &isEntry, 
				&retType, &itCalls, defby);
  cdbGetSourceWithinFileAttributes(object, &contSource, &sChar, &eChar);

  if (retType == NULL)
    pcdFcn = "Pcd";
  else
    pcdFcn = "Fct";

  if (displayMode == _dsplong)
    printf("RoutineSource: %s, ReturnType:%s, Start char: %d, End char: %d\n",
	 routineName, pcdFcn, sChar, eChar);
  else  if (displayMode == _xcdbmode)
    printf("%s %d %s %d %d",
	 routineName, object->objectKey(), defby, sChar, eChar);
  else
    printf("%d %d %c %s\n", sChar, eChar, pcdFcn[0], routineName);
}   /* objectRS() */


void
objectFLD(Field *object)
// --------------------------------------------------
{
  char objectName[MAX_STR_SIZE];
  RecordTypeDef *isMemberOf;

  cdbGetFieldAttributes( object, &isMemberOf);
  cdbGetDefinitionAttributes(object, objectName);

  if (displayMode == _dsplong)
    printf("Field: %s\n", objectName);
  else
    printf("%s", objectName);
}   /* objectFLD() */


void
objectDD(DatumDef *object)
// --------------------------------------------------
{
  char objectName[MAX_STR_SIZE];
  bool isConst;
  bool isForArg;
  int sChar;
  int eChar;
  SourceUnit *contSource;

  cdbGetDatumDefAttributes( object, &isConst, &isForArg);
  cdbGetDefinitionAttributes(object, objectName);
  cdbGetStandaloneDefinitionAttributes(object, &sChar, &eChar, &contSource);

  if (displayMode == _dsplong)
    printf("DatumDef: %s, startChar: %d, endChar: %d\n", 
	   objectName, sChar, eChar);
  else
    printf("%d %d %s", sChar, eChar, objectName);
}   /* objectDD() */


void
objectEC(EnumerationConst *object)
// --------------------------------------------------
{
  char objectName[MAX_STR_SIZE];
  EnumeratedTypeDef *isEnum;

  cdbGetEnumerationConstAttributes( object, &isEnum);
  cdbGetDefinitionAttributes(object, objectName);

  if (displayMode == _dsplong)
    printf("EnumerationConst: %s\n", objectName);
  else
    printf("%s",objectName);
}   /* objectEC() */


void
objectETD(EnumeratedTypeDef *object)
// --------------------------------------------------
{
  char objectName[MAX_STR_SIZE];
  int sChar;
  int eChar;
  SourceUnit *contSource;
  VoidIterator<SoftwareObject*> *itEnumMem;

  cdbGetEnumeratedTypeDefAttributes( object, &itEnumMem);
  cdbGetDefinitionAttributes(object, objectName);
  cdbGetStandaloneDefinitionAttributes(object, &sChar, &eChar, &contSource);

  if (displayMode == _dsplong)
    printf("EnumeratedTypeDef: %s, startChar: %d, endChar: %d\n", 
	   objectName, sChar, eChar);
  else
    printf("%d %d %s", sChar, eChar, objectName);
}   /* objectETD() */


void
objectRTD(RecordTypeDef *object)
// --------------------------------------------------
{
  char objectName[MAX_STR_SIZE];
  int sChar;
  int eChar;
  SourceUnit *contSource;
  VoidIterator<SoftwareObject*> *itFldMem;

  cdbGetRecordTypeDefAttributes( object, &itFldMem);
  cdbGetDefinitionAttributes(object, objectName);
  cdbGetStandaloneDefinitionAttributes(object, &sChar, &eChar, &contSource);

  if (displayMode == _dsplong)
    printf("RecordTypeDef: %s, startChar: %d, endChar: %d\n", 
	   objectName, sChar, eChar);
  else
    printf("%d %d %s", sChar, eChar, objectName);
}   /* objectRTD() */


void
objectMCE(ManifestConstExistence *object)
// --------------------------------------------------
{
  char objectName[MAX_STR_SIZE];
  VoidIterator<SoftwareObject*> *itSource,
	*itsucCompound,
	*itpreCompound;

  cdbDataUseExistenceAttributes(object, &itSource, &itsucCompound,
				&itpreCompound);
  cdbGetReferenceExistenceAttributes(object, objectName);

  if (displayMode == _dsplong) 
      printf("ManifestConstExistence: %s\n", objectName);
  else
    printf("%s", objectName);
}   /* objectMCE() */


void
objectDUE(DataUseExistence *object)
// --------------------------------------------------
{
  char objectName[MAX_STR_SIZE];
  VoidIterator<SoftwareObject*> *itSource,
	*itsucCompound,
	*itpreCompound;

  cdbDataUseExistenceAttributes(object, &itSource, &itsucCompound,
				&itpreCompound);
  cdbGetReferenceExistenceAttributes(object, objectName);

  if (displayMode == _dsplong) 
      printf("DataUseExistence: %s\n", objectName);
  else
    printf("%s", objectName);
}   /* objectDUE() */


void
objectTUE(TypeUseExistence *object)
// --------------------------------------------------
{
  char objectName[MAX_STR_SIZE];
  VoidIterator<SoftwareObject*> *itType,
	*itretType;

  cdbTypeUseExistenceAttributes(object, &itType, &itretType);
  cdbGetReferenceExistenceAttributes(object, objectName);

  if (displayMode == _dsplong)
    printf("TypeUseExistence: %s\n", objectName);
  else
    printf("%s", objectName);
}   /* objectTUE() */


void
objectRCE(RoutineCallExistence *object)
// --------------------------------------------------
{
  char objectName[MAX_STR_SIZE];
  VoidIterator<SoftwareObject*> *itCalledBy, *defList;
  SoftwareObject *rsObj;


  cdbRoutineCallExistenceAttributes(object, &itCalledBy);
  cdbGetReferenceExistenceAttributes(object, objectName);

    if (displayMode == _xcdbmode) {
    defList = cdbGetDefList(object);
    if (!defList->eoi() ) {
      rsObj = defList->next(); 
      objectRS((RoutineSource *)rsObj);
   }
    else
      printf("%s", objectName);
 
 }


  else  if (displayMode == _dsplong)
    printf("RoutineCallExistence: %s\n", objectName);
  else
    printf("%s", objectName);
}   /* objectRCE() */


void
objectFIE(FileInclusionExistence *object)
// --------------------------------------------------
{
  char objectName[MAX_STR_SIZE];
  VoidIterator<SoftwareObject*> *itIncludedIn, *defList;
  SoftwareObject *sfileObj;

  cdbGetFileInclusionExistenceAttributes(object, &itIncludedIn);
  cdbGetReferenceExistenceAttributes(object, objectName);
  
  
  if (displayMode == _xcdbmode) {
     defList = cdbGetDefList(object);
     if (!defList->eoi() ) {
      sfileObj = defList->next(); 
      objectSF((SourceFile *)sfileObj);
    }
     else
      printf("%s", objectName);
 }

  else if (displayMode == _dsplong)
    printf("FileInclusionExistence: %s\n", objectName);
  else
    printf("%s", objectName);
}   /* objectFIE() */


/*****************************************************
            SoftwareObject -> Qelement
 *****************************************************/


Qelement *
SO2QeltIterator::computeNext() 
{
  SoftwareObject *nextSO;
  Qelement *currentQelt;

  currentQelt = NULL;
  while ( (currentQelt == NULL) && (! _iterationBase->eoi()) ) {
    nextSO = _iterationBase->next();

    if (compatibleSO( _headSO, nextSO)) {
      currentQelt = new Qelement(_headSO);
      currentQelt->merge(nextSO);
      _headDealtWith = true;
      _eoi = _iterationBase->eoi();
    }
    else {
      if (! _headDealtWith)
        currentQelt = new Qelement(_headSO);
      _headSO = nextSO;
      _headDealtWith = false;
    }
  }

  if (currentQelt == NULL) {
    currentQelt = new Qelement(_headSO);
    _eoi = true;
  }

  return(currentQelt);
}   // SO2QeltIterator::computeNext()

 
 	
	


/*****************************************************

 *****************************************************/

int
SO_key_cmp(SoftwareObject *x, SoftwareObject *y)
//--------------------------------------------------
// compares objects on the basis of keys only

{

  if ( x->objectKey() < y->objectKey() )
    return( -1);
  else
    if ( x->objectKey() > y->objectKey() )
      return( 1); 
    else 
      return( 0);
  
}  /* SO_key_cmp */


VoidIterator<SoftwareObject*> *associationQuery(VoidIterator<Qelement*> *QiterIn,
						enum association_t rel,
						class_t filter)
// --------------------------------------------------
{

  VoidIterator<Qelement*> *Qiter;
  Qelement *tmpQ;
  VoidIterator<SoftwareObject*>
	*SOiterObj, 	// a SO iterator for each obj in `SOiterQel'
	*SOiterRtn;     // a SO iterator for gathering the results

  if (rel == _thatref )
    Qiter = new useViewIterator(QiterIn);
  else 
    Qiter = new defViewIterator(QiterIn);
  SOiterRtn = new VoidIterator<SoftwareObject*>();
  while (! Qiter->eoi()) {
    tmpQ = Qiter->next();
      switch (rel) {
      case _thatdef:
	SOiterObj=cdbGetObjectListThatDefineObject( filter,tmpQ->getDefObject());
	break;
      case _defby:
	SOiterObj=cdbGetObjectListDefinedByObject(  filter,tmpQ->getDefObject());
	break;
      case _thatref:
	SOiterObj=cdbGetObjectListThatReferToObject(filter,tmpQ->getUseObject());
	break;
      case _refby:
	SOiterObj=cdbGetObjectListReferedByObject(  filter,tmpQ->getDefObject());
	break;
      }
  
      SOiterRtn = new SortedUnionIterator<SoftwareObject*>(
					  SOiterObj, SOiterRtn, SO_key_cmp, SOmerge);


  }  // while (! Qiter->eoi())

  return(SOiterRtn);
}   /* associationQuery() */


void
loadDatabase(char *dbName)
// --------------------------------------------------
{
  char msg[MAX_STR_SIZE];
  int openDBres;

  /* closing the old one */
  if (the_DB != NULL)   
   terminateFileMem();

   the_DB = NULL;  // we just close the DB -> set the global variable to NULL

  /* opening a new database */
  openDBres = cdbInitReadOnlyFileMem( dbName);

  switch (openDBres) {
  case 0:
    database.DB_Name = dupString(dbName);
    init_pointers_from_db();
    break;

  case UFM_ALREADY_INITIALIZED:
    sprintf(msg, "UFM ALREADY INITIALIZED %s (rc=%d)", dbName, openDBres);
    errorMsg( msg);
    break;

  case UFM_FILE_NOT_SPECIFIED: 
    sprintf(msg, "warning: no database specified (rc=%d)", openDBres);
    errorMsg( msg);
    break;
  
  case UFM_FILE_NOT_FOUND: 
    sprintf(msg, "error: file not found %s (rc=%d)", dbName, openDBres);
    errorMsg( msg);
    break;

  default:
    sprintf(msg, "error: failed to init %s (rc=%d)", dbName, openDBres);
    errorMsg( msg);
  }  //switch 

}   /* loadDatabase() */
 


Qelement *
defViewIterator::computeNext()
  //--------------------------------------------------
  // Put _nextComputed into a local var 'tmp' and then 
  // call computeOneAhead() which will either put a new
  // value in _nextComputed or set it NULL and _eoi true
  // This returns 'tmp' in any case

  {
    Qelement *tmp;                  // for return object   
 
    tmp = _nextComputed;            // set up return object
    computeOneAhead();              // sets _nextComputed or _eoi
    return(tmp);

  };   /* defViewIterator::computeNext() */

void
defViewIterator::computeOneAhead()
  //--------------------------------------------------
  // Compute the next value for _nextComputed if possible, 
  // otherwise set _eoi to true.  If some objects are queued
  // in the variable _tempQueue, use the first queued item. 
  // Otherwise, see if next input item qualifies as is (it 
  // has a Definition component) 
  // If not, see if it can generate qualifying items, store 
  // any extras in _tempQueue if necessary.
                                             
{
  Qelement *nextQ_In;
  Qelement *tempQelt;
  VoidIterator<SoftwareObject *> *tempSOiter;

  _nextComputed = NULL;            // initialize  _nextComputed   
  if (! _tempQueue->eoi())
    /* there is something already queued..  we use it */
    _nextComputed = _tempQueue->next();
  else
    /* nothing is queued,  look for a Definition object to use */
    while ((_nextComputed == NULL) && (! _iterationBase->eoi())) {
      nextQ_In = _iterationBase->next();
      if (nextQ_In->defKey() != NULLKEY)
	/* if this input element has a Definition component,  use it */
	  _nextComputed = nextQ_In; 
      else {
	/* otherwise, see if it is associated with Definition objects */
	_tempQueue = new SO2QeltIterator( nextQ_In->
			  getObjectList(virtualizeclass(nextQ_In->qClass())) );
	if (! _tempQueue->eoi() ) {
	  /* if we got something, see if the first item has a defKey  */
	  tempQelt = _tempQueue->next();
	  /* if a defKey is present, use this Qelement */
	  if (tempQelt->defKey() != NULLKEY )
	    _nextComputed = tempQelt;
	}
      }
    }
  /* if there is nothing left, set _eoi to true */
  _eoi = ( _nextComputed == NULL);  
} // defViewIterator::computeNext()

Qelement *
useViewIterator::computeNext()
  //--------------------------------------------------
  // Put _nextComputed into a local var 'tmp' and then 
  // call computeOneAhead() which will either put a new
  // value in _nextComputed or set it NULL and _eoi true
  // This returns 'tmp'  in any case

  {
    Qelement *tmp;                  // for return object   
 
    tmp = _nextComputed;            // set up return object
    computeOneAhead();              // sets _nextComputed or _eoi
    return(tmp);

  };   /* useViewIterator::computeNext() */


void
useViewIterator::computeOneAhead()
  //-----------------------------------------------------
  // Objects are found in the input (or are derived from the 
  // input) that qualify as an output (they have a UseExistence
  // component, distinct from the previous output item).
  // A qualifying item is found and recorded in 
  // _computeOneAhead  if possible.  Otherwise, set _eoi to true.
  //
  // If next input item qualifies for the output as is,
  // use it.  Otherwise see if there is an associated object
  // which qualifies, provided we have not already done this.

{
  Qelement *nextQ_In;      // for holding the next input Qelement
  VoidIterator<SoftwareObject*> *tempSOiter; // a temporary 'list' holder
  SoftwareObject *tmpSO;  // a temporary SoftwareObject holder
  bool found;  // a flag for indicating when an output has been determined

  found = false;
  while ((! found) && (! _iterationBase->eoi() ) ) {
    // While input elements remain and output has not yet been computed
    nextQ_In = _iterationBase->next();
    if ((_nextComputed == NULL) || ( ! nextQ_In->compatible( _nextComputed) )) {
      // either this is the first pass or the next input elementis distinct anyway 
      if ( nextQ_In->useKey() != NULLKEY) { 
	// a useKey component is already present in this element, just use it.
	_nextComputed = nextQ_In;
	found = true;
      }
      else { // no useKey component is present in this element
	tempSOiter = nextQ_In->getObjectList(virtualizeclass(nextQ_In->qClass())); 
	// there is guaranteed at least one object, check if it is a useExistence
	tmpSO = tempSOiter->next();
	if (tmpSO->is_a(_referenceexistence) )  {
	  // the first item is the useExistence we sought,  
	  nextQ_In->merge(tmpSO);  // merge it with the Qelt
	  _nextComputed = nextQ_In;  // this is the next output Qelement
	  found = true;
	} // if useKey component found
	
      } // else (no useKey was initially present)
    } // if this input element was not compatible with previous
  } // while loop
  
  // if no new output is computed and the input list is finished, set _eoi true
  _eoi = ( (! found) && ( _iterationBase->eoi() )  );



}  // useViewIterator::computeOneAhead() 




