//                               -*- Mode: C++ -*- 
// api.hh  ---     : API between query language and DB
// Author          : Javed Chowdhury.
// Created On      : 
//

# ifndef API_H
# define API_H

extern "C" {
# include "c-util.h"
}

# include "Q-element.hh"

// types for the %union in YACC
enum association_t { _thatdef, _defby, _thatref, _refby };
enum operator_t    { _and, _or, _minus };

void sysinfo();
void printDBinfo();
void classinfo(class_t);
void objinfo(int);
int  classObjCount(class_t);

// OBJINFO related functions
void objectSF( SourceFile*);
void objectRS( RoutineSource*);
void objectMCE(ManifestConstExistence*);
void objectDUE(DataUseExistence*);
void objectTUE(TypeUseExistence*);
void objectFLD(Field*);
void objectDD( DatumDef*);
void objectEC( EnumerationConst*);
void objectETD(EnumeratedTypeDef*);
void objectRTD(RecordTypeDef*);
void objectRCE(RoutineCallExistence*);
void objectFIE(FileInclusionExistence*);

VoidIterator<Qelement*> *getObjectListRegExp( class_t, char*);
VoidIterator<SoftwareObject*> *associationQuery( VoidIterator<Qelement*>*,
												 enum association_t, class_t);

void loadDatabase (char*);


class SO2QeltIterator: public CastIterator<SoftwareObject*,Qelement*> {
protected:
  SoftwareObject *_headSO;  
  bool _headDealtWith;
  
  Qelement *computeNext();

public:
  SO2QeltIterator(VoidIterator<SoftwareObject*> *base)
    : CastIterator<SoftwareObject*,Qelement*>(base) {
      _headDealtWith = false;
      if (! _iterationBase->eoi()) {
        _headSO = _iterationBase->next();
        _eoi = false;
      }
      else
        _eoi = true;
  }
};  // class SO2QeltIterator


class defViewIterator: 
   public BaseIterator<VoidIterator<Qelement*> *, Qelement *> {

protected:

Qelement *_nextComputed;  // to buffer the next output element
VoidIterator<Qelement*> *_tempQueue;     // a temp Iterator for Qelements
void  computeOneAhead();  // sets _nextComputed value
Qelement *computeNext();  // return _nextComputed and maybe
                          // call computeOneAhead()

public:
  defViewIterator(VoidIterator<Qelement*> *base)
    :BaseIterator<VoidIterator<Qelement*> *, Qelement *>(base) {

      _nextComputed = NULL;
      _tempQueue = new VoidIterator<Qelement*>();
      if (! _iterationBase->eoi()) 
	computeOneAhead();
      else 
	_eoi = true;
  }
};// class defViewIterator


class useViewIterator: 
   public BaseIterator<VoidIterator<Qelement*> *, Qelement *> {

protected:

Qelement *_nextComputed;  // to buffer the next output element
void  computeOneAhead();  // sets '_nextComputed' value or _eoi
Qelement *computeNext();  // return _nextComputed and 
                          // call computeOneAhead()

public:
     useViewIterator(VoidIterator<Qelement*> *base)
       : BaseIterator<VoidIterator<Qelement*> *,Qelement*>(base) {

	 _nextComputed = NULL;
	 if (! _iterationBase->eoi() )
	   computeOneAhead();
 
  }
}; // class useViewIterator

# endif





