/*                               -*- Mode: C -*- 
 * query.y  ---    : scanner for query.
 * Author          : Javed Chowdhury.
 * Created On      : 
 */

%{
# include "c-util.h"
  /*  # define YY_USER_ACTION { printf(yytext); } */
/* BECAREFUL: not exactly the same definition as in the YACC part.  That's
   why we define it manually instead of using the `%union' instruction */
typedef union {
  char   *str;
  int     nbr;
/* the following is only needed in the YACC part.  It requires C++ compiler */
/*   class_t clss;
 *   enum association_t       assoc;
 *   enum operator_t          oper;
 *   VoidIterator<Qelement*> *iter;
 */
} YYSTYPE;

# include "query.tab.h"

int yylineno = 1;

/* non block-reads  needed to use query through a pipe.
 * From flex manual, p. 14 */
# undef YY_INPUT
# define YY_INPUT(buf,result,max_size) { \
    int c = getc(yyin);                   \
    result = (c==EOF) ? YY_NULL : (buf[0]=c,1); }

%}

A	[aA]
B	[bB]
C	[cC]
D	[dD]
E	[eE]
F	[fF]
G	[gG]
H	[hH]
I	[iI]
J	[jJ]
K	[kK]
L	[lL]
M	[mM]
N	[nN]
O	[oO]
P	[pP]
Q	[qQ]
R	[rR]
S	[sS]
T	[tT]
U	[uU]
V	[vV]
W	[wW]
X	[xX]
Y	[yY]
Z	[zZ]

DIG                     [0-9]
LTR                     [a-zA-Z_]
SPECCHAR                [][!?"'^*.$\\]|-

LWSC                    [ \t\v\f]
FILENAME                [-.0-9a-zA-Z_]
%%

{A}{N}{D}                   { return(AND); }
{C}{L}{A}{S}{S}{I}{N}{F}{O} { return(CLASS_INFO); }
{D}{A}{T}{A}{B}{A}{S}{E}    { return(DATABASE); }
{D}{I}{S}{P}{L}{A}{Y}       { return(DISPLAY); }
{F}{I}{R}{S}{T}             { return(FIRST); }
{G}{R}{A}{M}{A}{R}          { return(GRAMAR); }
{H}{E}{L}{P}                { return(HELP); }
{I}{D}                      { return(ID); }
{L}{O}{N}{G}                { return(LONG); }
{L}{O}{N}{G}{E}{R}          { return(LONGER); }
{N}{A}{M}{E}                { return(NAME); }
{R}{E}{S}{T}		    { return(REST); }  	
{N}{E}{X}{T}                { return(NEXT); }
{O}{B}{J}{I}{N}{F}{O}       { return(OBJ_INFO); }
{O}{R}                      { return(OR); }
{P}{A}{T}                   { return(PAT); }
{Q}{U}{I}{T}                { return(QUIT); }
{R}{E}{G}                   { return(REG); }
{S}{E}{T}                   { return(SET); }
{S}{H}{O}{R}{T}             { return(SHORT); }
{S}{Y}{S}{I}{N}{F}{O}       { return(SYS_INFO); } 
{T}{O}{P}{N}                { return(TOPN); }
{X}{C}{D}{B}{M}{O}{D}{E}    { return(XCDBMODE); }

{C}{T}{E}                   { return(CTE); }
{D}{D}                      { return(DD); }
{D}{E}                      { return(DE); }
{D}{U}{E}                   { return(DUE); }
{E}{C}                      { return(EC); }
{E}{T}{D}                   { return(ETD); }
{F}{I}{E}                   { return(FIE); }
{F}{L}{D}                   { return(FLD); }
{M}{C}{E}                   { return(MCE); }
{R}{C}{E}                   { return(RCE); }
{R}{E}                      { return(RE); }
{R}{S}                      { return(RS); }
{R}{T}{D}                   { return(RTD); }
{S}{C}{E}                   { return(SCE); }
{S}{D}                      { return(SD); }
{S}{F}                      { return(SF); }
{S}{O}                      { return(SO); }
{S}{U}                      { return(SU); }
{S}{W}{F}                   { return(SWF); }
{T}{D}                      { return(TD); }
{T}{D}{D}                   { return(TDD); }
{T}{U}{E}                   { return(TUE); }

{D}{A}{T}                   { return(DAT); }
{F}{I}{L}                   { return(FIL); }
{R}{O}{U}                   { return(ROU); }
{T}{Y}{P}                   { return(TYP); }

{D}{E}{F}{B}{Y}             { return(DEFBY); }
{R}{E}{F}{B}{Y}             { return(REFBY); }
{T}{H}{A}{T}{D}{E}{F}       { return(THATDEF); }
{T}{H}{A}{T}{R}{E}{F}       { return(THATREF); }

"-"                         { return(MINUS); }
"="                         { return(EQUAL); }
"("                         { return(LEFT_PARENT); }
")"                         { return(RIGHT_PARENT); }

{DIG}+                      { yylval.nbr = atoi(yytext);
                              return(NUM); }

({LTR}|{DIG}|{SPECCHAR})+   { yylval.str = dupString(yytext);
                              return(IDENTIFIER); } 


("/"?|{FILENAME})+          { yylval.str = dupString(yytext);
                              return(FNAME); }

\n                          { yylineno++;
                              return(EOL);}
{LWSC}                       /* Absorb whitespace character */
.                            /* Trap any non-standard characters */

<<EOF>>                     { return(END_OF_FILE); }

%%


int yyerror(char *s)
/************************************************************
 */
{ 
  char msg[MAX_STR_SIZE];
  
  sprintf(msg, "%s: at or before '%s' line %d\n", s, yytext, yylineno);
  errorMsg( msg);
}
