/*                               -*- Mode: C -*-
 * query.y  ---    : grammar for query.
 * Author          : Javed Chowdhury.
 * Created On      :
 * Modified by Tim Lethbridge to call optimized Reg Exp Search
 */

%{
# include <stdlib.h>
# include "Q-element.hh"
# include "optimizing.hh"
# include "write-API.hh"
# include "read-API.hh"
# include "utilities.hh"

# define DEFAULT_DB   "test.mf"
# define HELPFILE     "README.QL"
# define THISFILE     "query.y"


/* defined in the scanner which is compiled in C (see Makefile) */
extern "C" int yylex();
extern "C" int yyerror(char*);

static VoidIterator<Qelement*> *queryIter = NULL;

/* BECAREFUL: not exactly the same definition as in the LEX part.  That's
   why we define it manually instead of using the `%union' instruction */
typedef union {
  char   *str;
  int     nbr;
  /* the following is not in the LEX part because it requires C++ compiler */
  class_t clss;
  enum association_t       assoc;
  enum operator_t          oper;
  VoidIterator<Qelement*> *iter;
} YYSTYPE;
%}


%token CLASS_INFO DATABASE DISPLAY END_OF_FILE EOL EQUAL FIRST GRAMAR REST XCDBMODE
%token HELP ID LEFT_PARENT LONG LONGER NAME NEXT OBJ_INFO PAT QUIT REG RIGHT_PARENT
%token SET SHORT SYS_INFO TOPN

%token FIL ROU DAT TYP DD DE DUE EC ETD FIE FLD MCE RCE RE RS RTD CTE
%token SCE SD SF SO SU SWF TD TDD TUE

%token <str> FNAME IDENTIFIER

%token <nbr> NUM

%type <str> pattern regexpr
%type <clss> class regularclass specialclass
%type <assoc> association
%type <oper> operator
%type <iter> query

%left AND MINUS OR
%left THATDEF DEFBY THATREF REFBY

%%

session
   : expr_list QUIT EOL    { YYACCEPT; }
   | expr_list END_OF_FILE { YYACCEPT; }
   ;

expr_list
   :
   | expr_list expr EOL   { prompt(); }
   | error EOL            { yyerrok; prompt(); }
   ;

expr
   :                      /* empty query */
   | query  { 
               if (queryIter != NULL) {
		 delete( queryIter);
		 queryIter = NULL;
	       }
	       if (the_DB != NULL) 
		 printQuery( $1); 
   }
   | FIRST query  {
               if (the_DB != NULL)  { 
		 if (queryIter != NULL)  {
		   delete( queryIter);
		   queryIter = NULL;
		 }
		 printQueryN( $2, topN);
		 queryIter = $2; 
	       } 
   } 

   | FIRST NUM query  {
	       if (the_DB != NULL) { 
		 if (queryIter != NULL)  {
		   delete( queryIter);
		   queryIter = NULL;
		 }
		 printQueryN( $3, $2);
		 queryIter = $3;
	       }  
   } 
   | NEXT    {
	       if (the_DB == NULL)
		 errorMsg(NODB_MSG); 
	       else { 
		 if (queryIter != NULL)
		   printQueryN( queryIter, topN);
	       } 
   }
   | NEXT NUM  {
	       if (the_DB == NULL) 
		 errorMsg(NODB_MSG); 
	       else { 
		 if (queryIter != NULL)
		   printQueryN( queryIter, $2); 
	       }
   } 

   | REST    {
	       if (the_DB == NULL) 
		 errorMsg(NODB_MSG); 
	       else { 
		 if (queryIter != NULL)
		   printQuery( queryIter); 
	       }
   }

   | OBJ_INFO NUM    { 
               if (the_DB == NULL) 
		 errorMsg(NODB_MSG); 
	       else {
		 objinfo($2);
		 printf("\n");
		 fflush(stdout);
		 } 
   }
   | CLASS_INFO class    { 
               if (the_DB == NULL) 
		 errorMsg(NODB_MSG); 
	       else
		 classinfo($2); 
   }
   | SYS_INFO           { 
               if (the_DB == NULL) 
		 errorMsg(NODB_MSG); 
	       else
		 sysinfo();
   }
   | SET set_expr
   | HELP               { printHelp(HELPFILE); }
   | GRAMAR             { printHelp(THISFILE); }
   ;

query
   : ID NUM  {
	       if (the_DB == NULL) 
		 errorMsg(NODB_MSG); 
	       else { 
		 if ( ($2 >= cdbGetObjectCount(_softwareobject)) ||
					($2 < 0) )   {
		   errorMsg("wrong object key !");
		   $$ = new VoidIterator<Qelement*>();
		 }
		 else
		   $$ = new SO2QeltIterator(cdbGetObjectListGivenKey($2)); 
	       }
   }
   | NAME regexpr {
	       if (the_DB == NULL) 
		 errorMsg(NODB_MSG); 
	       else { 
		 $$ = new SO2QeltIterator(
                         cdbGetObjectListGivenName($2, _softwareobject));
		 free($2); 
	       } 
   }
   | class NAME regexpr {
               if (the_DB == NULL) 
		 errorMsg(NODB_MSG); 
	       else { 
		 $$ = new SO2QeltIterator(cdbGetObjectListGivenName( $3, $1));
		 free($3); 
	       }
   }
   | REG regexpr {
	       if (the_DB == NULL) 
		 errorMsg(NODB_MSG); 
	       else { 
		 /*printf("REG EXP:`%s'\n", $2); */
		 $$ = new SO2QeltIterator(
                      cdbGetObjectListOptRegExp( $2, _softwareobject));
		 free($2); 
	       }
   }
   | class REG regexpr {
	       if (the_DB == NULL) 
		 errorMsg(NODB_MSG); 
	       else { 
		 $$ = new SO2QeltIterator(
                      cdbGetObjectListOptRegExp( $3, $1));
		 free($3); 
	       }
   }
   | PAT pattern {
	       if (the_DB == NULL) 
		 errorMsg(NODB_MSG); 
	       else { 
		 $$ = new SO2QeltIterator(
			cdbGetObjectListOptRegExp( $2, _softwareobject));
		 free($2); 
	       }
   }
   | class PAT pattern {
	       if (the_DB == NULL) 
		 errorMsg(NODB_MSG); 
	       else { 
		 $$ =  new SO2QeltIterator(
                      cdbGetObjectListOptRegExp( $3, $1));
		 free($3); 
	       }
   }
   | association query        %prec THATDEF {
               if (the_DB == NULL) 
		 errorMsg(NODB_MSG); 
	       else { 
		 $$ = new SO2QeltIterator(
		      associationQuery( $2, $1, _softwareobject)); 
	       }
   }
   | class association query  %prec THATDEF {
	       if (the_DB == NULL) 
		 errorMsg(NODB_MSG); 
	       else { 
		 $$ = new SO2QeltIterator(
		      associationQuery( $3, $2, $1)); 
	       }
   }
   | query operator query     %prec AND   {
	       if (the_DB == NULL) 
		 errorMsg(NODB_MSG); 
	       else { 
		 switch ($2)  {
		 case _and:
		   $$=new SortedInterIterator<Qelement*>( $1, $3,
			    qeltCmp, qeltMerge);
		   break;
		 case _or:
		   $$=new SortedUnionIterator<Qelement*>( $1, $3,
			    qeltCmp, qeltMerge);
		   break;
		 case _minus:
		   $$=new SortedMinusIterator<Qelement*>( $1, $3, qeltCmp);
		   break;
		 }
	       }
   }
   | LEFT_PARENT query RIGHT_PARENT  { $$ = $2; } 
   ;

set_expr
   : DATABASE EQUAL FNAME      { loadDatabase( $3); }
   | DATABASE EQUAL IDENTIFIER { loadDatabase( $3); }
   | TOPN EQUAL NUM            { topN = $3; }
   | DISPLAY EQUAL SHORT       { displayMode = _dspshort; }
   | DISPLAY EQUAL LONG        { displayMode = _dsplong; }
   | DISPLAY EQUAL LONGER      { displayMode = _dsplonger; }
   | DISPLAY EQUAL XCDBMODE    { displayMode = _xcdbmode; printf("xcdbmode\n"); fflush(stdout); }
    ;

class
   : regularclass
   | specialclass
   ;

regularclass
   : SO     { $$ = _softwareobject; }
   | CTE    { $$ = _commenttermexistence; }
   | DD     { $$ = _datumdef; }
   | DE     { $$ = _definition; }
   | DUE    { $$ = _datauseexistence; }
   | EC     { $$ = _enumerationconst; }
   | ETD    { $$ = _enumeratedtypedef; }
   | FIE    { $$ = _fileinclusionexistence; }
   | FLD    { $$ = _field; }
   | MCE    { $$ = _manifestconstexistence; }
   | RCE    { $$ = _routinecallexistence; }
   | RE     { $$ = _referenceexistence; }
   | RS     { $$ = _routinesource; }
   | RTD    { $$ = _recordtypedef; }
   | SCE    { $$ = _specialcodeexistence; }
   | SD     { $$ = _standalonedefinition; }
   | SF     { $$ = _sourcefile; }
   | SU     { $$ = _sourceunit; }
   | SWF    { $$ = _sourcewithinfile; }
   | TD     { $$ = _typedef; }
   | TDD    { $$ = _typeddefinition; }
   | TUE    { $$ = _typeuseexistence; }
   ;

specialclass
   : ROU      { $$ = _routine; }
   | FIL      { $$ = _file; }
   | DAT      { $$ = _data; }
   | TYP      { $$ = _type; }
   ;

association
   : THATDEF  { $$ = _thatdef; }
   | DEFBY    { $$ = _defby; }
   | THATREF  { $$ = _thatref; }
   | REFBY    { $$ = _refby; }
   ;

operator
   : AND      { $$ = _and; }
   | OR       { $$ = _or; }
   | MINUS    { $$ = _minus; }
   ;

pattern
   : regexpr    { $$ = patternToRegexpr( $1); free( $1); }
   ;

regexpr
   : IDENTIFIER
   | CTE        { $$ = dupString("cte"); }
   | DD         { $$ = dupString("dd"); }
   | DE         { $$ = dupString("de"); }
   | DUE        { $$ = dupString("due"); }
   | EC         { $$ = dupString("ec"); }
   | ETD        { $$ = dupString("etd"); }
   | FIE        { $$ = dupString("fie"); }
   | FLD        { $$ = dupString("fld"); }
   | MCE        { $$ = dupString("mce"); }
   | RCE        { $$ = dupString("rce"); }
   | RE         { $$ = dupString("re"); }
   | RS         { $$ = dupString("rs"); }
   | RTD        { $$ = dupString("rtd"); }
   | SCE        { $$ = dupString("sce"); }
   | SD         { $$ = dupString("sd"); }
   | SF         { $$ = dupString("sf"); }
   | SO         { $$ = dupString("so"); }
   | SU         { $$ = dupString("su"); }
   | SWF        { $$ = dupString("swf"); }
   | TD         { $$ = dupString("td"); }
   | TDD        { $$ = dupString("tdd"); }
   | TUE        { $$ = dupString("tue"); }

   | DAT        { $$ = dupString("dat"); }
   | FIL        { $$ = dupString("fil"); }
   | ROU        { $$ = dupString("rou"); }
   | TYP        { $$ = dupString("typ"); }

   | DEFBY      { $$ = dupString("defby"); }
   | REFBY      { $$ = dupString("refby"); }
   | THATDEF    { $$ = dupString("thatdef"); }
   | THATREF    { $$ = dupString("thatref"); }

   | AND        { $$ = dupString("and"); }
   | DATABASE   { $$ = dupString("database"); }
   | DISPLAY    { $$ = dupString("display"); }
   | FIRST      { $$ = dupString("first"); }
   | FNAME      { $$ = dupString("fname"); }
   | GRAMAR     { $$ = dupString("gramar"); }
   | HELP       { $$ = dupString("help"); }
   | ID         { $$ = dupString("id"); }
   | LONG       { $$ = dupString("long"); }
   | LONGER     { $$ = dupString("longer"); }
   | NAME       { $$ = dupString("name"); }
   | NEXT       { $$ = dupString("next"); }
   | NUM        { $$ = dupString("num"); }
   | OR         { $$ = dupString("or"); }
   | PAT        { $$ = dupString("pat"); }
   | QUIT       { $$ = dupString("quit"); }
   | REST       { $$ = dupString("rest"); }
   | REG        { $$ = dupString("reg"); }
   | SET        { $$ = dupString("set"); }
   | SHORT      { $$ = dupString("short"); }
   | TOPN       { $$ = dupString("topn"); }
   | XCDBMODE      { $$ = dupString("xcdbmode"); }
    ;
%%

/* Functions. */
main(int argc, char *argv[])
/**************************************************
 */
{
  setopt(argc, argv);
  
  loadDatabase(dupString(database.DB_Name));
  
  prompt();

  yyparse();

  /* closing */
  terminateFileMem();
}
