/*                               -*- Mode: C++ -*- 
 * utilities.cc    : utilities for query
 * Author          : Javed Chowdhury.
 * Created On      : 
 */

# include "utilities.hh"


void prompt()
/**************************************************
 */
{
  if (displayMode == _xcdbmode)
	printf(NO_PROMPT);
  else
	printf(OPQPROMPT);
  fflush(stdout);
}


void printHelp(char *helpfile)
/**************************************************
 */
{
  char cmd[MAX_STR_SIZE];

  sprintf(cmd,"$PAGER %s\n", helpfile);

  system(cmd);
}

void printQuery( VoidIterator<Qelement*> *iter)
/**************************************************
 */
{
  while ( ! iter->eoi()) 
    iter->next()->print();
}
  
  void printQueryN( VoidIterator<Qelement*> *iter, int topn)
/**************************************************
 */
{
  for ( ; (! iter->eoi()) && ( topn > 0); topn--)
    iter->next()->print();
}


void printObject( SoftwareObject *object)
/**************************************************
 */
{
  if (displayMode == _dsplong) {
    printf("Object Name: %s ", cdbGetObjectName( object));
    printf("Class: %s ", cdbClass2str( cdbGetClassGivenObject( object)));
    printf("Key: %d ", cdbGetKeyGivenObject( object));
  }
  else {
    printf("%s ", cdbGetObjectName( object));
    printf("%s ", cdbClass2abbrev( cdbGetClassGivenObject( object)));
    printf("%d ", cdbGetKeyGivenObject( object));
  }

  printf("\n");
}


char *patternToRegexpr( char *patt)
/**************************************************
 * converts a pattern into a reg-expr:
 * - adds ^ at the beginning of the pattern
 * - adds $ at the end of the pattern
 * - replaces ? by .
 * - replaces . by \.
 * - replaces * by .*
 */
{
  char *regexpr;
  int i, j;

  /* maximum possible size for the reg-exp: twice the length of the
  /* pattern + 1 (for the '\0') */

  regexpr = (char *)malloc((strlen(patt))*2+1);
  *regexpr = '^';

  for (i=0, j=0; patt[i] != '\0'; i++)
    switch (patt[i]) {
    case '*':
      regexpr[++j] = '.';
      regexpr[++j] = '*';
      break;
    case '.':
      regexpr[++j] = '\\';
      regexpr[++j] = '.';
      break;
    case '?':
      regexpr[++j] = '.';
      break;
    default:
      regexpr[++j] = patt[i];
      break;
    }
  regexpr[++j] = '$';
  regexpr[++j] = '\0';
  return(regexpr);
}


void unknowopt( char *opt, char *cmd, char *known)
/**************************************************
 */
{
  char msg[MAX_STR_SIZE];

  sprintf( msg, "unknown option :%s\n", opt);
  errorMsg( msg);
  sprintf( msg, "usage: %s %s\n", cmd, known);
  errorMsg( msg);
}

void setopt( int argc, char *argv[])
/**************************************************
 * parameters analysis
 */
{
  char *valid = "[-h?] [-b <database>]";  /* valid options: h & ? are for help */
  char *stropt;
  int   i;

  database.DB_Name = NULL;                /* initialize  DB_name member of database */

  for ( i=1; i < argc; i++) {
    stropt = argv[i];

    if (*stropt != '-')
      unknowopt( stropt, argv[0], valid);
    else
      switch ( *(stropt+1) ) {
      case 'b':                   /* database */
	i++;
	if (i >= argc) {
	  errorMsg("error: option '-b' requires a parameter <filename>");
	  fprintf( stderr, "usage: %s %s\n", argv[0], valid);  /* use errorMsg() here ? */
	  exit(2);  /* exit program with an error (non 0) code */
	}
	database.DB_Name = dupString(argv[i]);
	break;
      case 'h':                   /* help */
      case '?':                   /* help */
	fprintf( stderr, "usage: %s %s\n", argv[0], valid);
	fprintf( stderr, "\t[h]\thelp: this message\n");
	fprintf( stderr, "\t[?]\thelp: this message\n");
	fprintf( stderr, "\t[b <database>]: give a DB to open\n");
	exit(0);
	break;
      otherwise:
	unknowopt( stropt, argv[0], valid);
      }
  }
}
